@@ -2,6 +2,8 @@ import sys
 import time
 import os
 import atexit
+import threading
+from typing import Any
 
 from vertx import EventBus
 from queue import Queue
@@ -13,6 +15,7 @@ class BPMService:
     delay = 10
     call_timeout = 30.0
     handlers = []
+    address_dict : dict
 
     def __init__(self, host="localhost", port=7000, options=None, err_handler=None, ssl_context=None):
         if options is None:
@@ -23,17 +26,13 @@ class BPMService:
                                     ssl_context=ssl_context)
         self.setupServices()
         self.connect()
+        self.address_dict = {}
         atexit.register(self.close)
 
     def setupServices(self):
         pass
 
-    def send(self, address, headers=None, body=None):
-        ret = Queue()
-        self.eb_calls.send(address=address, headers=headers, body=body, reply_handler=lambda msg: ret.put(msg))
-        return ret.get(True, self.call_timeout)
-
-    def reply(self, body=None, address=None, headers=None):
+    def setAddress(self, address: Any = None):
         if isinstance(address, dict):
             addr = address['replyAddress']
         elif isinstance(address, str):
@@ -42,12 +41,34 @@ class BPMService:
             addr = os.environ['BPM_EVENT_BUS_REPLY']
         else:
             raise ValueError('No address supplied')
-        if not isinstance(body,dict):
+        tid = threading.get_ident()
+        self.address_dict[tid]=addr
+
+    def getAddress(self, address: Any = None) -> str:
+        if isinstance(address, dict):
+            return address['replyAddress']
+        elif isinstance(address, str):
+            return address
+        elif threading.get_ident() in self.address_dict:
+            return self.address_dict[threading.get_ident()]
+        elif 'BPM_EVENT_BUS_REPLY' in os.environ:
+            return os.environ['BPM_EVENT_BUS_REPLY']
+        else:
+            raise ValueError('No address supplied')
+
+    def send(self, address, headers=None, body=None):
+        ret = Queue()
+        self.eb_calls.send(address=address, headers=headers, body=body, reply_handler=lambda msg: ret.put(msg))
+        return ret.get(True, self.call_timeout)
+
+    def reply(self, body=None, address=None, headers=None):
+        addr = self.getAddress(address)
+        if not isinstance(body, dict):
             body = {"reply": body}
         self.eb_calls.send(address=addr, headers=headers, body=body)
 
     def call(self, address, body=None, headers=None):
-        ret=self.send(address, headers, body)
+        ret = self.send(address, headers, body)
         if 'body' in ret:
             if 'reply' in ret['body']:
                 return ret['body']['reply']
@@ -81,13 +102,28 @@ class BPMService:
         self.eb_handlers.unregister_handler(address, handler)
         self.handlers.remove(address)
 
-    def remote_exec(self, msg: dict, script: str, lang: str = 'javascript', context: dict = None):
+    def run(self, address, handler):
+        self.register_handler(address, handler)
+        self.start()
+
+    def exec(self, script: str, lang: str = 'javascript', context: dict = None, address: object = None) -> object:
+        addr = self.getAddress(address)
         ret = Queue()
-        self.eb_calls.send(address=msg['body']['address'], body={
+        self.eb_calls.send(address=addr, body={
             "lang": lang,
             "script": script,
             "context": context
         }, reply_handler=lambda x: ret.put(x))
+        result = ret.get(True, self.call_timeout)
+        if result['body']['succeeded']:
+            return result['body']['body']
+        else:
+            raise ValueError(result['body']['trace'])
+
+    def request(self, request: str, address=None):
+        addr = self.getAddress(address)
+        ret = Queue()
+        self.eb_calls.send(address=addr, body={"request": request}, reply_handler=lambda x: ret.put(x))
         return ret.get(True, self.call_timeout)
 
     def start(self):
